"""Unit tests for MMO replace operation.
"""
import logging
import os

from deriva.core import mmo
from deriva.core.ermrest_model import tag
from tests.deriva.core.mmo.base import BaseMMOTestCase

logger = logging.getLogger(__name__)
logger.setLevel(os.getenv('DERIVA_PY_TEST_LOGLEVEL', default=logging.WARNING))


class TestMMOReplace (BaseMMOTestCase):

    def _pre(self, fn):
        """Pre-condition evaluation."""
        fn(self.assertTrue, self.assertFalse)

    def _post(self, fn):
        """Post-condition evaluation"""
        fn(self.assertFalse, self.assertTrue)

    def test_replace_col_in_vizcols(self):
        def cond(before, after):
            before(len(mmo.find(self.model, ["dept_schema", "dept", "postal_code"])) == 1)
            after(len(mmo.find(self.model, ["dept_schema", "dept", "zip"])) == 1)

        self._pre(cond)
        mmo.replace(self.model, ["dept_schema", "dept", "postal_code"], ["dept_schema", "dept", "zip"])
        self._post(cond)

    def test_replace_col_in_vizcols_pseudocol_simple(self):
        def cond(before, after):
            before(len(mmo.find(self.model, ["dept_schema", "dept", "street_address"])) == 1)
            after(len(mmo.find(self.model, ["dept_schema", "dept", "number_and_street_name"])) == 1)

        self._pre(cond)
        mmo.replace(self.model, ["dept_schema", "dept", "street_address"], ["dept_schema", "dept", "number_and_street_name"])
        self._post(cond)

    def test_replace_col_in_sourcedefs_columns(self):
        def cond(before, after):
            before(len(mmo.find(self.model, ["dept_schema", "dept", "country"])) == 1)
            after(len(mmo.find(self.model, ["dept_schema", "dept", "country_code"])) == 1)

        self._pre(cond)
        mmo.replace(self.model, ["dept_schema", "dept", "country"], ["dept_schema", "dept", "country_code"])
        self._post(cond)

    def test_replace_col_in_vizcols_pseudocol(self):
        def cond(before, after):
            before(len(mmo.find(self.model, ["dept_schema", "dept", "state"])) == 1)
            after(len(mmo.find(self.model, ["dept_schema", "dept", "state_or_province"])) == 1)

        self._pre(cond)
        mmo.replace(self.model, ["dept_schema", "dept", "state"], ["dept_schema", "dept", "state_or_province"])
        self._post(cond)

    def test_replace_col_in_sourcedefs_sources(self):
        def cond(before, after):
            before(len(mmo.find(self.model, ["dept_schema", "dept", "city"])) == 1)
            after(len(mmo.find(self.model, ["dept_schema", "dept", "township"])) == 1)

        self._pre(cond)
        mmo.replace(self.model, ["dept_schema", "dept", "city"], ["dept_schema", "dept", "township"])
        self._post(cond)

    def _do_test_replace_key_in_vizcols(self, oldfk, newfk):
        replacement = newfk
        newfk = [replacement[0], replacement[1] if replacement[1] else oldfk[1]]

        def cond(before, after):
            before(len(mmo.find(self.model, oldfk)) == 1)
            after(len(mmo.find(self.model, newfk)) == 1)

        self._pre(cond)
        mmo.replace(self.model, oldfk, replacement)
        self._post(cond)

    def test_replace_key_in_vizcols(self):
        self._do_test_replace_key_in_vizcols(["dept_schema", "dept_RID_key"], ["dept_schema", "dept_RID_key1"])

    def test_replace_key_in_vizcols_by_schema_only(self):
        self._do_test_replace_key_in_vizcols(["dept_schema", "dept_RID_key"], ["new_schema", None])

    def _do_test_replace_fkey_in_vizsrc(self, tagname, oldfk, newfk):
        replacement = newfk
        newfk = [replacement[0], replacement[1] if replacement[1] else oldfk[1]]

        def cond(before, after):
            before(any([m.tag == tagname and m.mapping == oldfk for m in mmo.find(self.model, oldfk)]))
            after(any([m.tag == tagname and m.mapping == newfk for m in mmo.find(self.model, newfk)]))

        self._pre(cond)
        mmo.replace(self.model, oldfk, replacement)
        self._post(cond)

    def test_replace_fkey_in_vizfkeys(self):
        self._do_test_replace_fkey_in_vizsrc(tag.visible_foreign_keys, ["person_schema", "person_dept_fkey"], ["person_schema", "person_dept_fkey1"])

    def test_replace_fkey_in_vizcols(self):
        self._do_test_replace_fkey_in_vizsrc(tag.visible_columns, ["person_schema", "person_dept_fkey"], ["person_schema", "person_dept_fkey1"])

    def test_replace_fkey_in_sourcedefs_fkeys(self):
        self._do_test_replace_fkey_in_vizsrc(tag.source_definitions, ["person_schema", "person_dept_fkey"], ["person_schema", "person_dept_fkey1"])

    def test_replace_fkey_in_vizfkeys_by_schema_only(self):
        self._do_test_replace_fkey_in_vizsrc(tag.visible_foreign_keys, ["person_schema", "person_dept_fkey"], ["new_schema", None])

    def test_replace_fkey_in_vizcols_by_schema_only(self):
        self._do_test_replace_fkey_in_vizsrc(tag.visible_columns, ["person_schema", "person_dept_fkey"], ["new_schema", None])

    def test_replace_fkey_in_sourcedefs_fkeys_by_schema_only(self):
        self._do_test_replace_fkey_in_vizsrc(tag.source_definitions, ["person_schema", "person_dept_fkey"], ["new_schema", None])

    def _do_test_replace_fkey_in_pseudocolumn(self, oldfk, newfk):
        replacement = newfk
        newfk = [replacement[0], replacement[1] if replacement[1] else oldfk[1]]

        def cond(before, after):
            before(any([m.tag == tag.visible_columns and isinstance(m.mapping, dict) for m in mmo.find(self.model, oldfk)]))
            after(any([m.tag == tag.visible_columns and isinstance(m.mapping, dict) for m in mmo.find(self.model, newfk)]))

        self._pre(cond)
        mmo.replace(self.model, oldfk, replacement)
        self._post(cond)

    def test_replace_fkey_in_pseudocolumn(self):
        self._do_test_replace_fkey_in_pseudocolumn(["person_schema", "person_dept_fkey"], ["person_schema", "person_dept_fkey1"])

    def test_replace_fkey_in_pseudocolumn_by_schema_only(self):
        self._do_test_replace_fkey_in_pseudocolumn(["person_schema", "person_dept_fkey"], ["new_schema", None])

    def _do_test_replace_fkey_in_sourcedefs_sources(self, oldfk, newfk):
        replacement = newfk
        newfk = [replacement[0], replacement[1] if replacement[1] else oldfk[1]]

        def cond(before, after):
            before(any([m.tag == tag.source_definitions and m.mapping == 'personnel' for m in mmo.find(self.model, oldfk)]))
            after(any([m.tag == tag.source_definitions and m.mapping == 'personnel' for m in mmo.find(self.model, newfk)]))

        self._pre(cond)
        mmo.replace(self.model, oldfk, replacement)
        self._post(cond)

    def test_replace_fkey_in_sourcedefs_sources(self):
        self._do_test_replace_fkey_in_sourcedefs_sources(["person_schema", "person_dept_fkey"], ["person_schema", "person_dept_fkey1"])

    def test_replace_fkey_in_sourcedefs_sources_by_schema_only(self):
        self._do_test_replace_fkey_in_sourcedefs_sources(["person_schema", "person_dept_fkey"], ["new_schema", None])

    def test_replace_col_in_search_box(self):
        def cond(before, after):
            before(len(mmo.find(self.model, ["person_schema", "person", "last_name"])) == 1)
            after(len(mmo.find(self.model, ["person_schema", "person", "surname"])) == 1)

        self._pre(cond)
        mmo.replace(self.model, ["person_schema", "person", "last_name"], ["person_schema", "person", "surname"])
        self._post(cond)

    def test_replace_all_constraints_by_schema_only(self):
        oldfk = ["person_schema", "person_dept_fkey"]
        newfk = ["new_schema", "person_dept_fkey"]
        oldk = ["person_schema", "person_RID_key"]
        newk = ["new_schema", "person_RID_key"]

        def cond(before, after):
            for tag_name, mapping_before, mapping_after, old_constraint, new_constraint in [
                (tag.visible_columns, oldk, newk, oldk, newk),
                (tag.visible_foreign_keys, oldfk, newfk, oldfk, newfk),
                (tag.source_definitions, 'personnel', 'personnel', oldfk, newfk)
            ]:
                before(any([m.tag == tag_name and m.mapping == mapping_before for m in mmo.find(self.model, old_constraint)]))
                after(any([m.tag == tag_name and m.mapping == mapping_after for m in mmo.find(self.model, new_constraint)]))

        self._pre(cond)
        mmo.replace(self.model, ["person_schema", None], ["new_schema", None])
        self._post(cond)




"""Base class for MMO test cases.
"""
import logging
import os
import unittest

from deriva.core import DerivaServer, ErmrestCatalog, get_credential
from deriva.core.ermrest_model import Schema, Table, Column, Key, ForeignKey, tag, builtin_types

logger = logging.getLogger(__name__)
logger.setLevel(os.getenv('DERIVA_PY_TEST_LOGLEVEL', default=logging.WARNING))
ermrest_hostname = os.getenv('DERIVA_PY_TEST_HOSTNAME')
ermrest_catalog_id = os.getenv('DERIVA_PY_TEST_CATALOG')
catalog = None

# baseline annotation doc for `dept` table
dept_annotations = {
    tag.visible_columns: {
        "compact": [
            ["dept_schema", "dept_RID_key"],
            ["dept_schema", "dept_dept_no_key"],
            "name"
        ],
        "detailed": [
            "RID",
            "RCT",
            {
                "source": "RMT",
                "markdown_name": "Last Modified Time"
            },
            "dept_no",
            "name",
            {
                "source": "street_address",
                "markdown_name": "Number and Street Name"
            },
            "postal_code",
            {
                "sourcekey": "head_count",
                "markdown_name": "Head Count"
            },
            {
                "display": {
                    "wait_for": [
                        "personnel"
                    ],
                    "template_engine": "handlebars",
                    "markdown_pattern": "{{#each personnel}}{{{this.values.name}}}{{#unless @last}}, {{/unless}}{{/each}}."
                },
                "markdown_name": "Personnel"
            }
        ]
    },
    tag.visible_foreign_keys: {
        "*": [
            [
                "person_schema",
                "person_dept_fkey"
            ]
        ]
    },
    tag.source_definitions: {
        "columns": [
            "dept_no",
            "name",
            "RID",
            "country"
        ],
        "sources": {
            "personnel": {
                "source": [
                    {
                        "inbound": [
                            "person_schema",
                            "person_dept_fkey"
                        ]
                    },
                    "name"
                ]
            },
            "head_count": {
                "source": [
                    {
                        "inbound": [
                            "person_schema",
                            "person_dept_fkey"
                        ]
                    },
                    "RID"
                ],
                "entity": False,
                "aggregate": "cnt_d"
            }
        }
    }
}

# baseline annotation doc for `person` table
person_annotations = {
    tag.visible_columns: {
        "compact": [
            ["person_schema", "person_RID_key"],
            "name"
        ],
        "detailed": [
            "RID",
            "name",
            ["person_schema", "person_dept_fkey"],
            {
                "markdown_name": "Department Name",
                "source": [
                    {
                        "outbound": [
                            "person_schema",
                            "person_dept_fkey"
                        ]
                    },
                    "name"
                ],
                "entity": False
            },
            {
                "sourcekey": "dept_size",
                "markdown_name": "Department Size"
            },
            {
                "sourcekey": "dept_city",
                "markdown_name": "City"
            },
            {
                "markdown_name": "State or Province",
                "source": [
                    {
                        "outbound": [
                            "person_schema",
                            "person_dept_fkey"
                        ]
                    },
                    "state"
                ],
                "entity": False
            },
        ]
    },
    tag.source_definitions: {
        "columns": [
            "RID",
            "name",
            "dept"
        ],
        "fkeys": [
            ["person_schema", "person_dept_fkey"]
        ],
        "sources": {
            "dept_size": {
                "source": [
                    {
                        "outbound": [
                            "person_schema",
                            "person_dept_fkey"
                        ]
                    },
                    {
                        "inbound": [
                            "person_schema",
                            "person_dept_fkey"
                        ]
                    },
                    "RID"
                ],
                "entity": False,
                "aggregate": "cnt_d"
            },
            "dept_city":             {
                "markdown_name": "City",
                "source": [
                    {
                        "outbound": [
                            "person_schema",
                            "person_dept_fkey"
                        ]
                    },
                    "city"
                ],
                "entity": False
            }
        },
        "search-box": {
            "or": [
                {
                    "source": "last_name"
                }
            ]
        }
    }
}


@unittest.skipUnless(ermrest_hostname, 'ERMrest hostname not defined.')
class BaseMMOTestCase (unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        BaseMMOTestCase.setUpCatalog()

    @classmethod
    def setUpCatalog(cls):
        global catalog

        # create catalog
        server = DerivaServer('https', ermrest_hostname, credentials=get_credential(ermrest_hostname))
        if ermrest_catalog_id:
            logger.debug(f'Connecting to {ermrest_hostname}/ermrest/catalog/{ermrest_catalog_id}')
            catalog = server.connect_ermrest(ermrest_catalog_id)
        else:
            catalog = server.create_ermrest_catalog()
            logger.debug(f'Created {ermrest_hostname}/ermrest/catalog/{catalog.catalog_id}')

        # get the  model
        model = catalog.getCatalogModel()

        # drop all schemas (except 'public')
        for sname in [sname for sname in model.schemas if sname != 'public']:
            model.schemas[sname].drop(cascade=True)

        # recreate schemas
        for sname in ["dept_schema", "person_schema"]:
            model.create_schema(Schema.define(sname))

        # create `dept` table
        model.schemas["dept_schema"].create_table(
            Table.define(
                'dept',
                column_defs=[
                    Column.define('dept_no', builtin_types.int8),
                    Column.define('name', builtin_types.text),
                    Column.define('street_address', builtin_types.text),
                    Column.define('city', builtin_types.text),
                    Column.define('state', builtin_types.text),
                    Column.define('country', builtin_types.text),
                    Column.define('postal_code', builtin_types.int8)
                ],
                key_defs=[
                    Key.define(['dept_no'])
                ],
                annotations=dept_annotations
            )
        )

        # create `person` table
        model.schemas["person_schema"].create_table(
            Table.define(
                'person',
                column_defs=[
                    Column.define('name', builtin_types.text),
                    Column.define('dept', builtin_types.int8),
                    Column.define('last_name', builtin_types.text)
                ],
                fkey_defs=[
                    ForeignKey.define(['dept'], "dept_schema", 'dept', ['dept_no'])
                ],
                annotations=person_annotations
            )
        )

        # populate for good measure (though not necessary for current set of tests)
        pbuilder = catalog.getPathBuilder()

        pbuilder.dept_schema.dept.insert([
            {'dept_no': 1, 'name': 'Dept A', 'street_address': '123 Main St', 'city': 'Anywhere', 'state': 'CA', 'country': 'US', 'postal_code': 98765},
            {'dept_no': 2, 'name': 'Dept B', 'street_address': '777 Oak Ave', 'city': 'Somewhere', 'state': 'NY', 'country': 'US', 'postal_code': 12345}
        ])

        pbuilder.person_schema.person.insert([
            {'name': 'John', 'dept': 1},
            {'name': 'Helena', 'dept': 1},
            {'name': 'Ben', 'dept': 1},
            {'name': 'Sonia', 'dept': 2},
            {'name': 'Rafael', 'dept': 2},
        ])

    @classmethod
    def tearDownClass(cls):
        BaseMMOTestCase.tearDownCatalog()

    @classmethod
    def tearDownCatalog(cls):
        global catalog
        if not ermrest_catalog_id and isinstance(catalog, ErmrestCatalog) and int(catalog.catalog_id) > 1000:
            # note: the '... > 1000' clause is intended to safeguard against accidental deletion of production catalogs in the usual (lower) range
            catalog.delete_ermrest_catalog(really=True)
        catalog = None

    def setUp(self):
        # reset annotations to baseline
        assert isinstance(catalog, ErmrestCatalog)
        self.model = catalog.getCatalogModel()

    def tearDown(self):
        pass



"""Unit tests for MMO prune operation.
"""
import logging
import os

from deriva.core import mmo
from deriva.core.ermrest_model import tag
from tests.deriva.core.mmo.base import BaseMMOTestCase

logger = logging.getLogger(__name__)
logger.setLevel(os.getenv('DERIVA_PY_TEST_LOGLEVEL', default=logging.WARNING))


class TestMMOPrune (BaseMMOTestCase):

    def _pre(self, fn):
        """Pre-condition evaluation."""
        fn(self.assertTrue)

    def _post(self, fn):
        """Post-condition evaluation"""
        fn(self.assertFalse)

    def test_prune_key_in_vizcols(self):
        fkname = ["dept_schema", "dept_RID_key"]

        def cond(assertion):
            matches = mmo.find(self.model, fkname)
            assertion(len(matches) == 1)

        self._pre(cond)
        mmo.prune(self.model, fkname)
        self._post(cond)

    def test_prune_key_in_vizcols_by_schema_only(self):
        fkname = ["dept_schema", "dept_RID_key"]

        def cond(assertion):
            matches = mmo.find(self.model, fkname)
            assertion(len(matches) == 1)

        self._pre(cond)
        mmo.prune(self.model, [fkname[0], None])
        self._post(cond)

    def test_prune_col_in_vizcols(self):
        def cond(assertion):
            matches = mmo.find(self.model, ["dept_schema", "dept", "RCT"])
            assertion(len(matches) == 1)

        self._pre(cond)
        mmo.prune(self.model, ["dept_schema", "dept", "RCT"])
        self._post(cond)

    def test_prune_col_in_vizcols_pseudocol_simple(self):
        def cond(assertion):
            matches = mmo.find(self.model, ["dept_schema", "dept", "RMT"])
            assertion(len(matches) == 1)

        self._pre(cond)
        mmo.prune(self.model, ["dept_schema", "dept", "RMT"])
        self._post(cond)

    def test_prune_col_in_vizcols_pseudocol(self):
        def cond(assertion):
            matches = mmo.find(self.model, ["dept_schema", "dept", "name"])
            assertion(any([m.anchor.name == 'person' and m.tag == tag.visible_columns and isinstance(m.mapping, dict) for m in matches]))

        self._pre(cond)
        mmo.prune(self.model, ["dept_schema", "dept", "name"])
        self._post(cond)

    def test_prune_col_in_sourcedefs_columns(self):
        def cond(assertion):
            matches = mmo.find(self.model, ["person_schema", "person", "dept"])
            assertion(any([m.anchor.name == 'person' and m.tag == tag.source_definitions and m.mapping == 'dept' for m in matches]))

        self._pre(cond)
        mmo.prune(self.model, ["person_schema", "person", "dept"])
        self._post(cond)

    def test_prune_col_in_sourcedefs_sources(self):
        def cond(assertion):
            matches = mmo.find(self.model, ["person_schema", "person", "RID"])
            assertion(any([m.tag == tag.source_definitions and m.mapping == 'dept_size' for m in matches]))

        self._pre(cond)
        mmo.prune(self.model, ["person_schema", "person", "RID"])
        self._post(cond)

    def test_prune_fkey_in_vizfkeys(self):
        fkname = ["person_schema", "person_dept_fkey"]

        def cond(assertion):
            matches = mmo.find(self.model, fkname)
            assertion(any([m.tag == tag.visible_foreign_keys and m.mapping == fkname for m in matches]))

        self._pre(cond)
        mmo.prune(self.model, fkname)
        self._post(cond)

    def test_prune_fkey_in_vizfkeys_by_schema_only(self):
        fkname = ["person_schema", "person_dept_fkey"]

        def cond(assertion):
            matches = mmo.find(self.model, fkname)
            assertion(any([m.tag == tag.visible_foreign_keys and m.mapping == fkname for m in matches]))

        self._pre(cond)
        mmo.prune(self.model, [fkname[0], None])
        self._post(cond)

    def test_prune_fkey_in_vizcols(self):
        fkname = ["person_schema", "person_dept_fkey"]

        def cond(assertion):
            matches = mmo.find(self.model, fkname)
            assertion(any([m.tag == tag.visible_columns and m.mapping == fkname for m in matches]))

        self._pre(cond)
        mmo.prune(self.model, fkname)
        self._post(cond)

    def test_prune_fkey_in_vizcols_by_schema_only(self):
        fkname = ["person_schema", "person_dept_fkey"]

        def cond(assertion):
            matches = mmo.find(self.model, fkname)
            assertion(any([m.tag == tag.visible_columns and m.mapping == fkname for m in matches]))

        self._pre(cond)
        mmo.prune(self.model, [fkname[0], None])
        self._post(cond)

    def test_prune_fkey_in_sourcedefs_sources(self):
        fkname = ["person_schema", "person_dept_fkey"]

        def cond(assertion):
            matches = mmo.find(self.model, fkname)
            assertion(any([m.tag == tag.source_definitions and m.mapping == 'personnel' for m in matches]))

        self._pre(cond)
        mmo.prune(self.model, fkname)
        self._post(cond)

    def test_prune_fkey_in_sourcedefs_sources_by_schema_only(self):
        fkname = ["person_schema", "person_dept_fkey"]

        def cond(assertion):
            matches = mmo.find(self.model, fkname)
            assertion(any([m.tag == tag.source_definitions and m.mapping == 'personnel' for m in matches]))

        self._pre(cond)
        mmo.prune(self.model, [fkname[0], None])
        self._post(cond)

    def test_prune_fkey_in_sourcedefs_fkeys(self):
        fkname = ["person_schema", "person_dept_fkey"]

        def cond(assertion):
            matches = mmo.find(self.model, fkname)
            assertion(any([m.tag == tag.source_definitions and m.mapping == fkname for m in matches]))

        self._pre(cond)
        mmo.prune(self.model, fkname)
        self._post(cond)

    def test_prune_fkey_in_sourcedefs_fkeys_by_schema_only(self):
        fkname = ["person_schema", "person_dept_fkey"]

        def cond(assertion):
            matches = mmo.find(self.model, fkname)
            assertion(any([m.tag == tag.source_definitions and m.mapping == fkname for m in matches]))

        self._pre(cond)
        mmo.prune(self.model, [fkname[0], None])
        self._post(cond)

    def test_prune_fkey_in_sourcedefs_recurse(self):
        def cond(assertion):
            assertion(any([
                isinstance(vizcol, dict) and vizcol.get("sourcekey") == "dept_size"
                for vizcol in self.model.schemas['person_schema'].tables['person'].annotations[tag.visible_columns]['detailed']
            ]))

        self._pre(cond)
        mmo.prune(self.model, ["person_schema", "person_dept_fkey"])
        self._post(cond)

    def test_prune_col_in_search_box(self):
        def cond(assertion):
            assertion(len(mmo.find(self.model, ["person_schema", "person", "last_name"])) == 1)

        self._pre(cond)
        mmo.prune(self.model, ["person_schema", "person", "last_name"])
        self._post(cond)



"""Unit tests for MMO find operation.
"""
import logging
import os

from deriva.core import mmo
from deriva.core.ermrest_model import tag
from tests.deriva.core.mmo.base import BaseMMOTestCase

logger = logging.getLogger(__name__)
logger.setLevel(os.getenv('DERIVA_PY_TEST_LOGLEVEL', default=logging.WARNING))


def first_element_match(l1, l2):
    return isinstance(l1, list) and isinstance(l2, list) and \
        len(l1) == len(l2) and len(l1) > 1 and \
        l1[0] == l2[0]


class TestMMOFind (BaseMMOTestCase):

    def test_find_key_in_vizcols(self):
        matches = mmo.find(self.model, ["dept_schema", "dept_RID_key"])
        self.assertEqual(len(matches), 1)

    def test_find_key_in_vizcols_by_schema_only(self):
        matches = mmo.find(self.model, ["dept_schema", None])
        self.assertGreater(len(matches), 0)

    def test_find_col_in_vizcols(self):
        matches = mmo.find(self.model, ["dept_schema", "dept", "RCT"])
        self.assertEqual(len(matches), 1)

    def test_find_col_in_vizcols_pseudocol_simple(self):
        matches = mmo.find(self.model, ["dept_schema", "dept", "RMT"])
        self.assertEqual(len(matches), 1)

    def test_find_col_in_vizcols_pseudocol(self):
        matches = mmo.find(self.model, ["dept_schema", "dept", "name"])
        self.assertTrue(any([m.anchor.name == 'person' and m.tag == tag.visible_columns and isinstance(m.mapping, dict) for m in matches]))

    def test_find_col_in_sourcedefs_columns(self):
        matches = mmo.find(self.model, ["person_schema", "person", "dept"])
        self.assertTrue(any([m.anchor.name == 'person' and m.tag == tag.source_definitions and m.mapping == 'dept' for m in matches]))

    def test_find_col_in_sourcedefs_sources(self):
        matches = mmo.find(self.model, ["person_schema", "person", "RID"])
        self.assertTrue(any([m.tag == tag.source_definitions and m.mapping == 'dept_size' for m in matches]))

    def test_find_fkey_in_vizfkeys(self):
        fkname = ["person_schema", "person_dept_fkey"]
        matches = mmo.find(self.model, fkname)
        self.assertTrue(any([m.tag == tag.visible_foreign_keys and m.mapping == fkname for m in matches]))

    def test_find_fkey_in_vizfkeys_by_schema_only(self):
        fkname = ["person_schema", None]
        matches = mmo.find(self.model, fkname)
        self.assertTrue(any([m.tag == tag.visible_foreign_keys and first_element_match(m.mapping, fkname) for m in matches]))

    def test_find_fkey_in_vizcols(self):
        fkname = ["person_schema", "person_dept_fkey"]
        matches = mmo.find(self.model, fkname)
        self.assertTrue(any([m.tag == tag.visible_columns and m.mapping == fkname for m in matches]))

    def test_find_fkey_in_vizcols_by_schema_only(self):
        fkname = ["person_schema", None]
        matches = mmo.find(self.model, fkname)
        self.assertTrue(any([m.tag == tag.visible_columns and first_element_match(m.mapping, fkname) for m in matches]))

    def test_find_fkey_in_sourcedefs_sources(self):
        matches = mmo.find(self.model, ["person_schema", "person_dept_fkey"])
        self.assertTrue(any([m.tag == tag.source_definitions and m.mapping == 'personnel' for m in matches]))

    def test_find_fkey_in_sourcedefs_sources_by_schema_only(self):
        matches = mmo.find(self.model, ["person_schema", None])
        self.assertTrue(any([m.tag == tag.source_definitions and m.mapping == 'personnel' for m in matches]))

    def test_find_fkey_in_sourcedefs_fkeys(self):
        fkname = ["person_schema", "person_dept_fkey"]
        matches = mmo.find(self.model, fkname)
        self.assertTrue(any([m.tag == tag.source_definitions and m.mapping == fkname for m in matches]))

    def test_find_fkey_in_sourcedefs_fkeys_by_schema_only(self):
        fkname = ["person_schema", None]
        matches = mmo.find(self.model, fkname)
        self.assertTrue(any([m.tag == tag.source_definitions and m.context == 'fkeys' and first_element_match(m.mapping, fkname) for m in matches]))

    def test_find_col_in_search_box(self):
        matches = mmo.find(self.model, ["person_schema", "person", "last_name"])
        self.assertTrue(len(matches) == 1)



# Tests for the datapath module.
#
# Environment variables:
#  DERIVA_PY_TEST_HOSTNAME: hostname of the test server
#  DERIVA_PY_CATALOG: catalog identifier of the reusable test catalog (optional)
#  DERIVA_PY_TEST_CREDENTIAL: user credential, if none, it will attempt to get credential for given hostname (optional)
#  DERIVA_PY_TEST_VERBOSE: set for verbose logging output to stdout (optional)

import logging
import os
import unittest
from deriva.core import DerivaServer, get_credential, ermrest_model, tag

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
if os.getenv("DERIVA_PY_TEST_VERBOSE"):
    logger.setLevel(logging.DEBUG)
else:
    logger.setLevel(logging.INFO)

hostname = os.getenv("DERIVA_PY_TEST_HOSTNAME")


@unittest.skipUnless(hostname, "Test host not specified")
class ErmrestModelTests (unittest.TestCase):

    catalog = None

    @classmethod
    def _purgeCatalog(cls):
        model = cls.catalog.getCatalogModel()

        # exclude the 'public' schema
        schemas = [s for s in model.schemas.values() if s.name != 'public']

        # drop all fkeys
        for s in schemas:
            for t in s.tables.values():
                for fk in list(t.foreign_keys):
                    fk.drop()

        # drop all tables and schemas
        for s in list(schemas):
            for t in list(s.tables.values()):
                t.drop()
            s.drop()

    @classmethod
    def setUpClass(cls):
        credential = os.getenv("DERIVA_PY_TEST_CREDENTIAL") or get_credential(hostname)
        server = DerivaServer('https', hostname, credentials=credential)
        catalog_id = os.getenv("DERIVA_PY_TEST_CATALOG")
        if catalog_id is not None:
            logger.info(f"Reusing catalog {catalog_id} on host {hostname}")
            cls.catalog = server.connect_ermrest(catalog_id)
            cls._purgeCatalog()
        else:
            cls.catalog = server.create_ermrest_catalog()
            logger.info(f"Created catalog {cls.catalog.catalog_id} on host {hostname}")

    @classmethod
    def tearDownClass(cls):
        if cls.catalog and os.getenv("DERIVA_PY_TEST_CATALOG") is None:
            logger.info(f"Deleting catalog {cls.catalog.catalog_id} on host {hostname}")
            cls.catalog.delete_ermrest_catalog(really=True)

    def setUp(self):
        self.model = self.catalog.getCatalogModel()

    def tearDown(self):
        self._purgeCatalog()

    def _create_schema_with_fkeys(self):
        """Creates a simple schema of two tables with a fkey relationship from child to parent."""

        # build a single, low-level catalog /schema POST operation
        # should be (slightly) faster and avoids using the client APIs under test in this module
        schema_def = ermrest_model.Schema.define('schema_with_fkeys')
        schema_def["tables"] = {
            "parent": ermrest_model.Table.define(
                'parent',
                column_defs=[
                    ermrest_model.Column.define('id', ermrest_model.builtin_types.text),
                    ermrest_model.Column.define('id_extra', ermrest_model.builtin_types.text),
                ],
                key_defs=[
                    ermrest_model.Key.define(['id'], constraint_name='parent_id_key'),
                    ermrest_model.Key.define(['id', 'id_extra'], constraint_name='parent_compound_key'),
                ]
            ),
            "child": ermrest_model.Table.define(
                'child',
                column_defs=[
                    ermrest_model.Column.define('parent_id', ermrest_model.builtin_types.text),
                    ermrest_model.Column.define('parent_id_extra', ermrest_model.builtin_types.text),
                ],
                fkey_defs=[
                    ermrest_model.ForeignKey.define(
                        ['parent_id'], 'schema_with_fkeys', 'parent', ['id']
                    ),
                    ermrest_model.ForeignKey.define(
                        ['parent_id_extra', 'parent_id'], 'schema_with_fkeys', 'parent', ['id_extra', 'id']
                    )
                ]
            ),
        }
        self.catalog.post('/schema', json=[schema_def])
        # refresh the local state of the model
        self.model = self.catalog.getCatalogModel()

    def test_0a_schema_define_defaults(self):
        sname = "test_schema"
        sdef = ermrest_model.Schema.define(sname)
        self.assertEqual(sdef.get("schema_name"), sname)
        self.assertEqual(sdef.get("comment"), None)
        self.assertEqual(sdef.get("acls"), dict())
        self.assertEqual(sdef.get("annotations"), dict())

    _test_comment = "my comment"
    _test_acls = {
        "insert": ["a"],
        "update": ["b"],
    }
    _test_acl_bindings = {
    }
    _test_annotations = {"tag1": "value1"}

    def test_0b_schema_define_custom(self):
        sname = "test_schema"
        sdef = ermrest_model.Schema.define(sname, self._test_comment, self._test_acls, self._test_annotations)
        self.assertEqual(sdef.get("schema_name"), sname)
        self.assertEqual(sdef.get("comment"), self._test_comment)
        self.assertEqual(sdef.get("acls"), self._test_acls)
        self.assertEqual(sdef.get("annotations"), self._test_annotations)

    def test_1a_column_define_defaults(self):
        cname = "test_column"
        cdef = ermrest_model.Column.define(cname, ermrest_model.builtin_types.text)
        self.assertEqual(cdef.get("name"), cname)
        ctype = cdef.get("type")
        self.assertIsInstance(ctype, dict)
        self.assertEqual(ctype.get("typename"), "text")
        self.assertEqual(cdef.get("nullok"), True)
        self.assertEqual(cdef.get("default"), None)
        self.assertEqual(cdef.get("comment"), None)
        self.assertEqual(cdef.get("acls"), dict())
        self.assertEqual(cdef.get("acl_bindings"), dict())
        self.assertEqual(cdef.get("annotations"), dict())

    def test_1b_column_define_custom(self):
        cname = "test_column"
        nullok = False
        default = "my default"
        cdef = ermrest_model.Column.define(
            cname, ermrest_model.builtin_types.text, nullok, default,
            self._test_comment, self._test_acls, self._test_acl_bindings, self._test_annotations,
        )
        self.assertEqual(cdef.get("name"), cname)
        ctype = cdef.get("type")
        self.assertIsInstance(ctype, dict)
        self.assertEqual(ctype.get("typename"), "text")
        self.assertEqual(cdef.get("nullok"), nullok)
        self.assertEqual(cdef.get("default"), default)
        self.assertEqual(cdef.get("comment"), self._test_comment)
        self.assertEqual(cdef.get("acls"), self._test_acls)
        self.assertEqual(cdef.get("acl_bindings"), self._test_acl_bindings)
        self.assertEqual(cdef.get("annotations"), self._test_annotations)

    def test_1c_key_define_defaults(self):
        cnames = ["id"]
        kdef = ermrest_model.Key.define(cnames)
        self.assertEqual(kdef.get("unique_columns"), cnames)
        self.assertEqual(not kdef.get("names"), True)
        self.assertEqual(kdef.get("comment"), None)
        self.assertEqual(kdef.get("annotations"), dict())

    def test_1d_key_define_custom(self):
        cnames = ["id"]
        constraint_name = "my_constraint"
        kdef = ermrest_model.Key.define(cnames, None, self._test_comment, self._test_annotations, constraint_name)
        self.assertEqual(kdef.get("unique_columns"), cnames)
        self.assertIsInstance(kdef.get("names"), list)
        self.assertIsInstance(kdef.get("names")[0], list)
        self.assertEqual(kdef.get("names")[0][1], constraint_name)
        self.assertEqual(kdef.get("comment"), self._test_comment)
        self.assertEqual(kdef.get("annotations"), self._test_annotations)

    def test_1e_fkey_define_defaults(self):
        fk_colnames = ["fk1"]
        pk_sname = "pk_schema"
        pk_tname = "pk_table"
        pk_colnames = ["RID"]
        fkdef = ermrest_model.ForeignKey.define(
            fk_colnames,
            pk_sname,
            pk_tname,
            pk_colnames,
        )
        self.assertEqual([ c.get("column_name") for c in fkdef.get("foreign_key_columns") ], fk_colnames)
        self.assertEqual([ c.get("column_name") for c in fkdef.get("referenced_columns") ], pk_colnames)
        self.assertEqual(all([ c.get("schema_name") == pk_sname for c in fkdef.get("referenced_columns") ]), True)
        self.assertEqual(all([ c.get("table_name") == pk_tname for c in fkdef.get("referenced_columns") ]), True)
        self.assertEqual(fkdef.get("on_update"), "NO ACTION")
        self.assertEqual(fkdef.get("on_delete"), "NO ACTION")
        self.assertEqual(not fkdef.get("names"), True)
        self.assertEqual(fkdef.get("comment"), None)
        self.assertEqual(fkdef.get("acls"), dict())
        self.assertEqual(fkdef.get("acl_bindings"), dict())
        self.assertEqual(fkdef.get("annotations"), dict())

    def test_1f_fkey_define_custom(self):
        fk_colnames = ["fk1"]
        pk_sname = "pk_schema"
        pk_tname = "pk_table"
        pk_colnames = ["RID"]
        constraint_name = "my_constraint"
        action1 = "CASCADE"
        action2 = "RESTRICT"
        fkdef = ermrest_model.ForeignKey.define(
            fk_colnames,
            pk_sname,
            pk_tname,
            pk_colnames,
            action1,
            action2,
            None,
            self._test_comment,
            self._test_acls,
            self._test_acl_bindings,
            self._test_annotations,
            constraint_name,
        )
        self.assertEqual([ c.get("column_name") for c in fkdef.get("foreign_key_columns") ], fk_colnames)
        self.assertEqual([ c.get("column_name") for c in fkdef.get("referenced_columns") ], pk_colnames)
        self.assertEqual(all([ c.get("schema_name") == pk_sname for c in fkdef.get("referenced_columns") ]), True)
        self.assertEqual(all([ c.get("table_name") == pk_tname for c in fkdef.get("referenced_columns") ]), True)
        self.assertEqual(fkdef.get("on_update"), action1)
        self.assertEqual(fkdef.get("on_delete"), action2)
        self.assertIsInstance(fkdef.get("names"), list)
        self.assertIsInstance(fkdef.get("names")[0], list)
        self.assertEqual(fkdef.get("names")[0][1], constraint_name)
        self.assertEqual(fkdef.get("comment"), self._test_comment)
        self.assertEqual(fkdef.get("acls"), self._test_acls)
        self.assertEqual(fkdef.get("acl_bindings"), self._test_acl_bindings)
        self.assertEqual(fkdef.get("annotations"), self._test_annotations)

    def test_2a_table_define_defaults(self):
        tname = "test_table"
        tdef = ermrest_model.Table.define(
            tname,
            [
                ermrest_model.Column.define("id", ermrest_model.builtin_types.text),
            ],
        )
        self.assertEqual(tdef.get("table_name"), tname)
        cdefs = tdef.get("column_definitions")
        self.assertIsInstance(cdefs, list)
        self.assertEqual(len(cdefs), 5 + 1)
        kdefs = tdef.get("keys")
        self.assertIsInstance(kdefs, list)
        self.assertEqual(len(kdefs), 1)
        fkdefs = tdef.get("foreign_keys")
        self.assertIsInstance(fkdefs, list)
        self.assertEqual(len(fkdefs), 2)
        self.assertEqual(tdef.get("comment"), None)
        self.assertEqual(tdef.get("acls"), dict())
        self.assertEqual(tdef.get("acl_bindings"), dict())
        self.assertEqual(tdef.get("annotations"), dict())
        schema = self.model.create_schema(ermrest_model.Schema.define("model_define_schema"))
        table = schema.create_table(tdef)
        self.assertIsInstance(table, ermrest_model.Table)

    def test_2b_table_define_custom(self):
        tname = "test_table"
        tdef = ermrest_model.Table.define(
            tname,
            [
                ermrest_model.Column.define("id", ermrest_model.builtin_types.text),
                ermrest_model.Column.define("fk1", ermrest_model.builtin_types.text),
            ],
            [
                ermrest_model.Key.define(["id"]),
            ],
            [
                ermrest_model.ForeignKey.define(["fk1"], "public", "ERMrest_Client", ["RID"]),
            ],
            self._test_comment,
            self._test_acls,
            self._test_acl_bindings,
            self._test_annotations,
            provide_system_fkeys=False,
        )
        self.assertEqual(tdef.get("table_name"), tname)
        cdefs = tdef.get("column_definitions")
        self.assertIsInstance(cdefs, list)
        self.assertEqual(len(cdefs), 5 + 2)
        kdefs = tdef.get("keys")
        self.assertIsInstance(kdefs, list)
        self.assertEqual(len(kdefs), 1 + 1)
        fkdefs = tdef.get("foreign_keys")
        self.assertIsInstance(fkdefs, list)
        self.assertEqual(len(fkdefs), 1)
        self.assertEqual(tdef.get("comment"), self._test_comment)
        self.assertEqual(tdef.get("acls"), self._test_acls)
        self.assertEqual(tdef.get("acl_bindings"), self._test_acl_bindings)
        self.assertEqual(tdef.get("annotations"), self._test_annotations)
        schema = self.model.create_schema(ermrest_model.Schema.define("model_define_schema"))
        table = schema.create_table(tdef)
        self.assertIsInstance(table, ermrest_model.Table)

    def test_2c_table_define_with_reference(self):
        tname = "test_table"
        tdef = ermrest_model.Table.define(
            tname,
            [
                ermrest_model.Column.define("id", ermrest_model.builtin_types.text),
                self.model.schemas["public"].tables["ERMrest_Client"],
            ],
            [],
            [],
            self._test_comment,
            self._test_acls,
            self._test_acl_bindings,
            self._test_annotations,
        )
        self.assertEqual(tdef.get("table_name"), tname)
        cdefs = tdef.get("column_definitions")
        self.assertIsInstance(cdefs, list)
        self.assertEqual(len(cdefs), 5 + 2)
        kdefs = tdef.get("keys")
        self.assertIsInstance(kdefs, list)
        self.assertEqual(len(kdefs), 1)
        fkdefs = tdef.get("foreign_keys")
        self.assertIsInstance(fkdefs, list)
        self.assertEqual(len(fkdefs), 2 + 1)
        self.assertEqual(
            {
                (
                    fkdef["referenced_columns"][0]["schema_name"],
                    fkdef["referenced_columns"][0]["table_name"],
                    frozenset(zip(
                        ( c["column_name"] for c in fkdef["foreign_key_columns"] ),
                        ( c["column_name"] for c in fkdef["referenced_columns"] ),
                    ))
                )
                for fkdef in fkdefs
             },
            {
                ("public", "ERMrest_Client", frozenset([ ("ERMrest_Client", "ID",) ])),
                ("public", "ERMrest_Client", frozenset([ ("RCB", "ID",) ])),
                ("public", "ERMrest_Client", frozenset([ ("RMB", "ID",) ])),
            }
        )
        self.assertEqual(tdef.get("comment"), self._test_comment)
        self.assertEqual(tdef.get("acls"), self._test_acls)
        self.assertEqual(tdef.get("acl_bindings"), self._test_acl_bindings)
        self.assertEqual(tdef.get("annotations"), self._test_annotations)
        schema = self.model.create_schema(ermrest_model.Schema.define("model_define_schema"))
        table = schema.create_table(tdef)
        self.assertIsInstance(table, ermrest_model.Table)

    def test_3a_vocab_define_defaults(self):
        tname = "test_table"
        curie_template = "test:{RID}"
        tdef = ermrest_model.Table.define_vocabulary(
            tname,
            curie_template,
        )
        self.assertEqual(tdef.get("table_name"), tname)
        cdefs = tdef.get("column_definitions")
        self.assertIsInstance(cdefs, list)
        self.assertEqual(len(cdefs), 5 + 5)
        self.assertEqual({ c["name"] for c in cdefs }, {"RID","RCT","RMT","RCB","RMB","ID","Name","Description","Synonyms","URI"})
        cdefs = { c["name"]: c for c in cdefs }
        self.assertEqual(cdefs["ID"]["default"], curie_template)
        self.assertEqual(cdefs["URI"]["default"], '/id/{RID}')
        kdefs = tdef.get("keys")
        self.assertIsInstance(kdefs, list)
        self.assertEqual(len(kdefs), 1 + 3)
        fkdefs = tdef.get("foreign_keys")
        self.assertIsInstance(fkdefs, list)
        self.assertEqual(len(fkdefs), 2)
        self.assertEqual(tdef.get("comment"), None)
        self.assertEqual(tdef.get("acls"), dict())
        self.assertEqual(tdef.get("acl_bindings"), dict())
        self.assertEqual(tdef.get("annotations"), dict())
        schema = self.model.create_schema(ermrest_model.Schema.define("model_define_schema"))
        table = schema.create_table(tdef)
        self.assertIsInstance(table, ermrest_model.Table)

    def test_3b_vocab_define_custom(self):
        tname = "test_table"
        curie_template = "test:{RID}"
        uri_template = "/id/1/{RID}"
        tdef = ermrest_model.Table.define_vocabulary(
            tname,
            curie_template,
            uri_template,
            [
                ermrest_model.Column.define("fk1", ermrest_model.builtin_types.text),
            ],
            [],
            [
                ermrest_model.ForeignKey.define(["fk1"], "public", "ERMrest_Client", ["RID"]),
            ],
            self._test_comment,
            self._test_acls,
            self._test_acl_bindings,
            self._test_annotations,
            provide_name_key=False,
            provide_system_fkeys=False,
        )
        self.assertEqual(tdef.get("table_name"), tname)
        cdefs = tdef.get("column_definitions")
        self.assertIsInstance(cdefs, list)
        self.assertEqual(len(cdefs), 5 + 5 + 1)
        self.assertEqual({ c["name"] for c in cdefs }, {"RID","RCT","RMT","RCB","RMB","ID","Name","Description","Synonyms","URI","fk1"})
        cdefs = { c["name"]: c for c in cdefs }
        self.assertEqual(cdefs["ID"]["default"], curie_template)
        self.assertEqual(cdefs["URI"]["default"], uri_template)
        kdefs = tdef.get("keys")
        self.assertIsInstance(kdefs, list)
        self.assertEqual(len(kdefs), 1 + 2)
        fkdefs = tdef.get("foreign_keys")
        self.assertIsInstance(fkdefs, list)
        self.assertEqual(len(fkdefs), 1)
        self.assertEqual(tdef.get("comment"), self._test_comment)
        self.assertEqual(tdef.get("acls"), self._test_acls)
        self.assertEqual(tdef.get("acl_bindings"), self._test_acl_bindings)
        self.assertEqual(tdef.get("annotations"), self._test_annotations)
        schema = self.model.create_schema(ermrest_model.Schema.define("model_define_schema"))
        table = schema.create_table(tdef)
        self.assertIsInstance(table, ermrest_model.Table)

    def test_3c_vocab_define_with_reference(self):
        tname = "test_table"
        curie_template = "test:{RID}"
        uri_template = "/id/1/{RID}"
        target = self.model.schemas["public"].tables["ERMrest_Client"]
        self.assertIsInstance(target, ermrest_model.Table)
        tdef = ermrest_model.Table.define_vocabulary(
            tname,
            curie_template,
            uri_template,
            [ target, ],
            [],
            [],
            self._test_comment,
            self._test_acls,
            self._test_acl_bindings,
            self._test_annotations,
        )
        self.assertEqual(tdef.get("table_name"), tname)
        cdefs = tdef.get("column_definitions")
        self.assertIsInstance(cdefs, list)
        self.assertEqual(len(cdefs), 5 + 5 + 1)
        self.assertEqual({ c["name"] for c in cdefs }, {"RID","RCT","RMT","RCB","RMB","ID","Name","Description","Synonyms","URI","ERMrest_Client"})
        kdefs = tdef.get("keys")
        self.assertIsInstance(kdefs, list)
        self.assertEqual(len(kdefs), 1 + 3)
        fkdefs = tdef.get("foreign_keys")
        self.assertIsInstance(fkdefs, list)
        self.assertEqual(len(fkdefs), 2 + 1)
        self.assertEqual(
            {
                (
                    fkdef["referenced_columns"][0]["schema_name"],
                    fkdef["referenced_columns"][0]["table_name"],
                    frozenset(zip(
                        ( c["column_name"] for c in fkdef["foreign_key_columns"] ),
                        ( c["column_name"] for c in fkdef["referenced_columns"] ),
                    ))
                )
                for fkdef in fkdefs
             },
            {
                ("public", "ERMrest_Client", frozenset([ ("ERMrest_Client", "ID",) ])),
                ("public", "ERMrest_Client", frozenset([ ("RCB", "ID",) ])),
                ("public", "ERMrest_Client", frozenset([ ("RMB", "ID",) ])),
            }
        )
        self.assertEqual(tdef.get("comment"), self._test_comment)
        self.assertEqual(tdef.get("acls"), self._test_acls)
        self.assertEqual(tdef.get("acl_bindings"), self._test_acl_bindings)
        self.assertEqual(tdef.get("annotations"), self._test_annotations)
        schema = self.model.create_schema(ermrest_model.Schema.define("model_define_schema"))
        table = schema.create_table(tdef)
        self.assertIsInstance(table, ermrest_model.Table)  

    def test_4a_asset_define_defaults(self):
        tname = "test_table"
        tdef = ermrest_model.Table.define_asset(
            "model_define_schema",
            tname,
        )
        self.assertEqual(tdef.get("table_name"), tname)
        cdefs = tdef.get("column_definitions")
        self.assertIsInstance(cdefs, list)
        self.assertEqual(len(cdefs), 5 + 5)
        self.assertEqual({ c["name"] for c in cdefs }, {"RID","RCT","RMT","RCB","RMB","Filename","Description","Length","MD5","URL"})
        cdefs = { c["name"]: c for c in cdefs }
        kdefs = tdef.get("keys")
        self.assertIsInstance(kdefs, list)
        self.assertEqual(len(kdefs), 1 + 1)
        fkdefs = tdef.get("foreign_keys")
        self.assertIsInstance(fkdefs, list)
        self.assertEqual(len(fkdefs), 2)
        self.assertEqual(tdef.get("acls"), dict())
        self.assertEqual(tdef.get("acl_bindings"), dict())
        self.assertEqual(set(tdef.get("annotations").keys()), {tag.table_display})
        schema = self.model.create_schema(ermrest_model.Schema.define("model_define_schema"))
        table = schema.create_table(tdef)
        self.assertIsInstance(table, ermrest_model.Table)

    def test_4b_asset_define_custom(self):
        tname = "test_table"
        hatrac_template = "/hatrac/foo/{{{MD5}}}.{{{Filename}}}"
        tdef = ermrest_model.Table.define_asset(
            "model_define_schema",
            tname,
            hatrac_template,
            [
                ermrest_model.Column.define("fk1", ermrest_model.builtin_types.text),
            ],
            [],
            [
                ermrest_model.ForeignKey.define(["fk1"], "public", "ERMrest_Client", ["RID"]),
            ],
            self._test_comment,
            self._test_acls,
            self._test_acl_bindings,
            self._test_annotations,
            provide_system_fkeys=False,
        )
        self.assertEqual(tdef.get("table_name"), tname)
        cdefs = tdef.get("column_definitions")
        self.assertIsInstance(cdefs, list)
        self.assertEqual(len(cdefs), 5 + 5 + 1)
        self.assertEqual({ c["name"] for c in cdefs }, {"RID","RCT","RMT","RCB","RMB","Filename","Description","Length","MD5","URL","fk1"})
        cdefs = { c["name"]: c for c in cdefs }
        self.assertEqual(cdefs["URL"]["annotations"][tag.asset]["url_pattern"], hatrac_template)
        kdefs = tdef.get("keys")
        self.assertIsInstance(kdefs, list)
        self.assertEqual(len(kdefs), 1 + 1)
        fkdefs = tdef.get("foreign_keys")
        self.assertIsInstance(fkdefs, list)
        self.assertEqual(len(fkdefs), 1)
        self.assertEqual(tdef.get("comment"), self._test_comment)
        self.assertEqual(tdef.get("acls"), self._test_acls)
        self.assertEqual(tdef.get("acl_bindings"), self._test_acl_bindings)
        annotations = tdef.get("annotations")
        for k, v in self._test_annotations.items():
            self.assertEqual(annotations[k], v)
        self.assertIn(tag.table_display, annotations)
        schema = self.model.create_schema(ermrest_model.Schema.define("model_define_schema"))
        table = schema.create_table(tdef)
        self.assertIsInstance(table, ermrest_model.Table)

    def test_4c_asset_define_with_reference(self):
        tname = "test_table"
        hatrac_template = "/hatrac/foo/{{{MD5}}}.{{{Filename}}}"
        target = self.model.schemas["public"].tables["ERMrest_Client"]
        self.assertIsInstance(target, ermrest_model.Table)
        tdef = ermrest_model.Table.define_asset(
            "model_define_schema",
            tname,
            hatrac_template,
            [ target, ],
            [],
            [],
            self._test_comment,
            self._test_acls,
            self._test_acl_bindings,
            self._test_annotations,
        )
        self.assertEqual(tdef.get("table_name"), tname)
        cdefs = tdef.get("column_definitions")
        self.assertIsInstance(cdefs, list)
        self.assertEqual(len(cdefs), 5 + 5 + 1)
        self.assertEqual({ c["name"] for c in cdefs }, {"RID","RCT","RMT","RCB","RMB","Filename","Description","Length","MD5","URL","ERMrest_Client"})
        kdefs = tdef.get("keys")
        self.assertIsInstance(kdefs, list)
        self.assertEqual(len(kdefs), 1 + 1)
        fkdefs = tdef.get("foreign_keys")
        self.assertIsInstance(fkdefs, list)
        self.assertEqual(len(fkdefs), 2 + 1)
        self.assertEqual(
            {
                (
                    fkdef["referenced_columns"][0]["schema_name"],
                    fkdef["referenced_columns"][0]["table_name"],
                    frozenset(zip(
                        ( c["column_name"] for c in fkdef["foreign_key_columns"] ),
                        ( c["column_name"] for c in fkdef["referenced_columns"] ),
                    ))
                )
                for fkdef in fkdefs
             },
            {
                ("public", "ERMrest_Client", frozenset([ ("ERMrest_Client", "ID",) ])),
                ("public", "ERMrest_Client", frozenset([ ("RCB", "ID",) ])),
                ("public", "ERMrest_Client", frozenset([ ("RMB", "ID",) ])),
            }
        )
        self.assertEqual(tdef.get("comment"), self._test_comment)
        self.assertEqual(tdef.get("acls"), self._test_acls)
        self.assertEqual(tdef.get("acl_bindings"), self._test_acl_bindings)
        annotations = tdef.get("annotations")
        for k, v in self._test_annotations.items():
            self.assertEqual(annotations[k], v)
        self.assertIn(tag.table_display, annotations)
        schema = self.model.create_schema(ermrest_model.Schema.define("model_define_schema"))
        table = schema.create_table(tdef)
        self.assertIsInstance(table, ermrest_model.Table)  

    def test_5a_assoctable_define_defaults(self):
        tdef = ermrest_model.Table.define_association(
            [
                self.model.schemas["public"].tables["ERMrest_Client"],
                self.model.schemas["public"].tables["ERMrest_Group"],
            ],
        )
        self.assertEqual(tdef.get("table_name"), "ERMrest_Client_ERMrest_Group")
        self.assertEqual(tdef.get("comment"), None)
        cdefs = tdef.get("column_definitions")
        self.assertIsInstance(cdefs, list)
        self.assertEqual(len(cdefs), 5 + 2)
        kdefs = tdef.get("keys")
        self.assertIsInstance(kdefs, list)
        self.assertEqual(
            { tuple(key["unique_columns"]) for key in kdefs },
            { ('RID',), ('ERMrest_Client', 'ERMrest_Group') },
        )
        fkdefs = tdef.get("foreign_keys")
        self.assertIsInstance(fkdefs, list)
        self.assertEqual(
            {
                (
                    fkdef["referenced_columns"][0]["schema_name"],
                    fkdef["referenced_columns"][0]["table_name"],
                    frozenset(zip(
                        tuple( c["column_name"] for c in fkdef["foreign_key_columns"] ),
                        tuple( c["column_name"] for c in fkdef["referenced_columns"] ),
                    ))
                )
                for fkdef in fkdefs
             },
            {
                ("public", "ERMrest_Client", frozenset([ ("ERMrest_Client", "ID",) ])),
                ("public", "ERMrest_Group", frozenset([ ("ERMrest_Group", "ID",) ])),
                ("public", "ERMrest_Client", frozenset([ ("RCB", "ID",) ])),
                ("public", "ERMrest_Client", frozenset([ ("RMB", "ID",) ])),
            }
        )
        schema = self.model.create_schema(ermrest_model.Schema.define("model_define_schema"))
        table = schema.create_table(tdef)
        self.assertIsInstance(table, ermrest_model.Table)

    def test_5b_assoctable_define_custom(self):
        tname = "client_group"
        tdef = ermrest_model.Table.define_association(
            [
                ('c_id', self.model.schemas["public"].tables["ERMrest_Client"]),
                ('g_id', self.model.schemas["public"].tables["ERMrest_Group"]),
            ],
            [],
            tname,
            self._test_comment,
            provide_system_fkeys=False,
        )
        self.assertEqual(tdef.get("table_name"), tname)
        self.assertEqual(tdef.get("comment"), self._test_comment)
        cdefs = tdef.get("column_definitions")
        self.assertIsInstance(cdefs, list)
        self.assertEqual(len(cdefs), 5 + 2)
        kdefs = tdef.get("keys")
        self.assertIsInstance(kdefs, list)
        self.assertEqual(
            { tuple(key["unique_columns"]) for key in kdefs },
            { ('RID',), ('c_id', 'g_id') },
        )
        fkdefs = tdef.get("foreign_keys")
        self.assertIsInstance(fkdefs, list)
        self.assertEqual(
            {
                (
                    fkdef["referenced_columns"][0]["schema_name"],
                    fkdef["referenced_columns"][0]["table_name"],
                    frozenset(zip(
                        tuple( c["column_name"] for c in fkdef["foreign_key_columns"] ),
                        tuple( c["column_name"] for c in fkdef["referenced_columns"] ),
                    ))
                )
                for fkdef in fkdefs
             },
            {
                ("public", "ERMrest_Client", frozenset([ ("c_id", "ID",) ])),
                ("public", "ERMrest_Group", frozenset([ ("g_id", "ID",) ])),
            }
        )
        schema = self.model.create_schema(ermrest_model.Schema.define("model_define_schema"))
        table = schema.create_table(tdef)
        self.assertIsInstance(table, ermrest_model.Table)

    def test_5c_assoctable_define_with_metadata(self):
        tname = "client_group"
        tdef = ermrest_model.Table.define_association(
            [
                ('c_id', self.model.schemas["public"].tables["ERMrest_Client"]),
                ('g_id', self.model.schemas["public"].tables["ERMrest_Group"]),
            ],
            [
                ermrest_model.Column.define("md1", ermrest_model.builtin_types.text),
            ],
            tname,
            self._test_comment,
        )
        self.assertEqual(tdef.get("table_name"), tname)
        self.assertEqual(tdef.get("comment"), self._test_comment)
        cdefs = tdef.get("column_definitions")
        self.assertIsInstance(cdefs, list)
        self.assertEqual(len(cdefs), 5 + 2 + 1)
        kdefs = tdef.get("keys")
        self.assertIsInstance(kdefs, list)
        self.assertEqual(
            { tuple(key["unique_columns"]) for key in kdefs },
            { ('RID',), ('c_id', 'g_id') },
        )
        fkdefs = tdef.get("foreign_keys")
        self.assertIsInstance(fkdefs, list)
        self.assertEqual(
            {
                (
                    fkdef["referenced_columns"][0]["schema_name"],
                    fkdef["referenced_columns"][0]["table_name"],
                    frozenset(zip(
                        tuple( c["column_name"] for c in fkdef["foreign_key_columns"] ),
                        tuple( c["column_name"] for c in fkdef["referenced_columns"] ),
                    ))
                )
                for fkdef in fkdefs
             },
            {
                ("public", "ERMrest_Client", frozenset([ ("c_id", "ID",) ])),
                ("public", "ERMrest_Group", frozenset([ ("g_id", "ID",) ])),
                ("public", "ERMrest_Client", frozenset([ ("RCB", "ID",) ])),
                ("public", "ERMrest_Client", frozenset([ ("RMB", "ID",) ])),
            }
        )
        schema = self.model.create_schema(ermrest_model.Schema.define("model_define_schema"))
        table = schema.create_table(tdef)
        self.assertIsInstance(table, ermrest_model.Table)

    def test_key_drop_cascading(self):
        self._create_schema_with_fkeys()
        schema = self.model.schemas['schema_with_fkeys']
        self.model.schemas['schema_with_fkeys'].tables['parent'].keys[(schema, 'parent_id_key')].drop(cascade=True)

    def test_key_reordered_drop_cascading(self):
        self._create_schema_with_fkeys()
        schema = self.model.schemas['schema_with_fkeys']
        self.model.schemas['schema_with_fkeys'].tables['parent'].keys[(schema, 'parent_compound_key')].drop(cascade=True)

    def test_key_column_drop_cascading(self):
        self._create_schema_with_fkeys()
        self.model.schemas['schema_with_fkeys'].tables['parent'].columns['id'].drop(cascade=True)

    def test_fkey_column_drop_cascading(self):
        self._create_schema_with_fkeys()
        self.model.schemas['schema_with_fkeys'].tables['child'].columns['parent_id_extra'].drop(cascade=True)

    def test_table_drop_cascading(self):
        self._create_schema_with_fkeys()
        self.model.schemas['schema_with_fkeys'].tables['parent'].drop(cascade=True)

    def test_schema_drop_cascading(self):
        self._create_schema_with_fkeys()
        self.model.schemas['schema_with_fkeys'].drop(cascade=True)


if __name__ == '__main__':
    unittest.main()



"""Unit tests for MMO+DDL Drop operations.
"""
import os
import logging
from deriva.core import mmo
from deriva.core.ermrest_model import UpdateMappings
from tests.deriva.core.mmo.base import BaseMMOTestCase

logger = logging.getLogger(__name__)
logger.setLevel(os.getenv('DERIVA_PY_TEST_LOGLEVEL', default=logging.WARNING))


class TestMMOxDDLDrop (BaseMMOTestCase):

    @classmethod
    def setUpClass(cls):
        """Don't bother setting up catalog for the suite.
        """
        pass

    def setUp(self):
        """Setup catalog for each unit test.
        """
        TestMMOxDDLDrop.setUpCatalog()
        super().setUp()

    def test_drop_key(self):
        kname = ["dept_schema", "dept_dept_no_key"]

        def test():
            matches = mmo.find(self.model, kname)
            return len(matches)

        self.assertTrue(test())
        t = self.model.schemas[kname[0]].tables["dept"]
        key = t.keys[(t.schema, kname[1])]
        key.drop(cascade=True, update_mappings=UpdateMappings.immediate)
        self.assertFalse(test())

    def test_drop_key_deferred(self):
        kname = ["dept_schema", "dept_dept_no_key"]

        def test(model):
            matches = mmo.find(model, kname)
            return len(matches)

        self.assertTrue(test(self.model))
        t = self.model.schemas[kname[0]].tables["dept"]
        key = t.keys[(t.schema, kname[1])]
        key.drop(cascade=True, update_mappings=UpdateMappings.deferred)
        self.assertTrue(test(self.model.catalog.getCatalogModel()))
        self.model.apply()
        self.assertFalse(test(self.model.catalog.getCatalogModel()))

    def test_drop_key_cascade(self):
        kname = ["dept_schema", "dept_dept_no_key"]
        fkname = ["person_schema", "person_dept_fkey"]

        def test(name):
            matches = mmo.find(self.model, name)
            return len(matches)

        self.assertTrue(test(kname))
        self.assertTrue(test(fkname))
        t = self.model.schemas[kname[0]].tables["dept"]
        key = t.keys[(t.schema, kname[1])]
        key.drop(cascade=True, update_mappings=UpdateMappings.immediate)
        self.assertFalse(test(kname))
        self.assertFalse(test(fkname))

    def test_drop_fkey(self):
        fkname = ["person_schema", "person_dept_fkey"]

        def test():
            matches = mmo.find(self.model, fkname)
            return len(matches)

        self.assertTrue(test())
        self.model.fkey(fkname).drop(update_mappings=UpdateMappings.immediate)
        self.assertFalse(test())

    def test_drop_fkey_deferred(self):
        fkname = ["person_schema", "person_dept_fkey"]

        def test(model):
            matches = mmo.find(model, fkname)
            return len(matches)

        self.assertTrue(test(self.model))
        self.model.fkey(fkname).drop(update_mappings=UpdateMappings.deferred)
        self.assertTrue(test(self.model.catalog.getCatalogModel()))
        self.model.apply()
        self.assertFalse(test(self.model.catalog.getCatalogModel()))

    def test_drop_col(self):
        cname = ["person_schema", "person", "last_name"]

        def test():
            matches = mmo.find(self.model, cname)
            return len(matches)

        self.assertTrue(test())
        t = self.model.schemas[cname[0]].tables[cname[1]]
        col = t.columns[cname[2]]
        col.drop(update_mappings=UpdateMappings.immediate)
        self.assertFalse(test())

    def test_drop_col_deferred(self):
        cname = ["person_schema", "person", "last_name"]

        def test(model):
            matches = mmo.find(model, cname)
            return len(matches)

        self.assertTrue(test(self.model))
        t = self.model.schemas[cname[0]].tables[cname[1]]
        col = t.columns[cname[2]]
        col.drop(update_mappings=UpdateMappings.deferred)
        self.assertTrue(test(self.model.catalog.getCatalogModel()))
        self.model.apply()
        self.assertFalse(test(self.model.catalog.getCatalogModel()))

    def test_drop_col_cascade_key(self):
        cname = ["dept_schema", "dept", "dept_no"]
        kname = ["dept_schema", "dept_dept_no_key"]
        fkname = ["person_schema", "person_dept_fkey"]

        def test(name):
            matches = mmo.find(self.model, name)
            return len(matches)

        for name in [cname, kname, fkname]:
            with self.subTest(f'subTest {name}'):
                self.assertTrue(test(name))

        t = self.model.schemas[cname[0]].tables[cname[1]]
        col = t.columns[cname[2]]
        col.drop(cascade=True, update_mappings=UpdateMappings.immediate)
        
        for name in [cname, kname, fkname]:
            with self.subTest(f'subTest {name}'):
                self.assertFalse(test(name))

    def test_drop_col_cascade_fkey(self):
        cname = ["person_schema", "person", "dept"]
        fkname = ["person_schema", "person_dept_fkey"]

        def test(name):
            matches = mmo.find(self.model, name)
            return len(matches)

        self.assertTrue(test(cname))
        self.assertTrue(test(fkname))
        t = self.model.schemas[cname[0]].tables[cname[1]]
        col = t.columns[cname[2]]
        col.drop(cascade=True, update_mappings=UpdateMappings.immediate)
        self.assertFalse(test(cname))
        self.assertFalse(test(fkname))

    def test_drop_table_cascade(self):
        cname = ["dept_schema", "dept", "dept_no"]
        kname = ["dept_schema", "dept_dept_no_key"]
        fkname = ["person_schema", "person_dept_fkey"]

        def test(name):
            matches = mmo.find(self.model, name)
            return len(matches)

        for name in [cname, kname, fkname]:
            with self.subTest(f'subTest {name}'):
                self.assertTrue(test(name))

        t = self.model.schemas[cname[0]].tables[cname[1]]
        t.drop(cascade=True, update_mappings=UpdateMappings.immediate)

        for name in [cname, kname, fkname]:
            with self.subTest(f'subTest {name}'):
                self.assertFalse(test(name))

    def test_drop_table_cascade_deferred(self):
        cname = ["person_schema", "person", "name"]
        kname = ["person_schema", "person_RID_key"]
        fkname = ["person_schema", "person_dept_fkey"]

        def test(model, name):
            matches = mmo.find(model, name)
            return len(matches)

        for name in [cname, kname, fkname]:
            with self.subTest(f'subtest {name} precondition'):
                self.assertTrue(test(self.model, name))

        t = self.model.schemas[cname[0]].tables[cname[1]]
        t.drop(cascade=True, update_mappings=UpdateMappings.deferred)

        # fkname should linger on in the dept table's annotations until the update is applied to the model
        self.assertTrue(test(self.model.catalog.getCatalogModel(), fkname))
        self.model.apply()
        self.assertFalse(test(self.model.catalog.getCatalogModel(), fkname))

    def test_drop_table(self):
        fkname = ["person_schema", "person_dept_fkey"]

        def test(name):
            matches = mmo.find(self.model, name)
            return len(matches)

        self.assertTrue(test(fkname))
        t = self.model.schemas["person_schema"].tables["person"]
        t.drop(update_mappings=UpdateMappings.immediate)
        self.assertFalse(test(fkname))

    def test_drop_schema_cascade(self):
        cname = ["dept_schema", "dept", "dept_no"]
        kname = ["dept_schema", "dept_dept_no_key"]
        fkname = ["person_schema", "person_dept_fkey"]

        def test(name):
            matches = mmo.find(self.model, name)
            return len(matches)

        for name in [cname, kname, fkname]:
            with self.subTest(f'subtest {name} precondition'):
                self.assertTrue(test(name))

        s = self.model.schemas[cname[0]]
        s.drop(cascade=True, update_mappings=UpdateMappings.immediate)

        for name in [cname, kname, fkname]:
            with self.subTest(f'subtest {name} postcondition'):
                self.assertFalse(test(name))

    def test_drop_schema_cascade_deferred(self):
        cname = ["person_schema", "person", "name"]
        kname = ["person_schema", "person_RID_key"]
        fkname = ["person_schema", "person_dept_fkey"]

        def test(model, name):
            matches = mmo.find(model, name)
            return len(matches)

        for name in [cname, kname, fkname]:
            with self.subTest(f'subtest {name} precondition'):
                self.assertTrue(test(self.model, name))

        s = self.model.schemas[cname[0]]
        s.drop(cascade=True, update_mappings=UpdateMappings.deferred)

        # fkname should linger on in the dept table's annotations until the update is applied to the model
        self.assertTrue(test(self.model.catalog.getCatalogModel(), fkname))
        self.model.apply()
        self.assertFalse(test(self.model.catalog.getCatalogModel(), fkname))



"""Model management operations.
"""
import logging
from collections import namedtuple

from deriva.core import tag as tags

logger = logging.getLogger(__name__)

Match = namedtuple('Match', 'anchor tag context container mapping')

__search_box__ = 'search-box'


def replace(model, symbol, replacement):
    """Replaces `symbol` with `replacement` where symbol is found in mappings.

    See the 'find' function for notes on what are 'symbols' and how symbols are found in the model.

    The `symbol` and its `replacement` must be of the same type (i.e., columns or constraints). For columns, only the
    column name may differ. For constraints, the whole name or just the schema part may be replaced.
    """
    logger.debug(f'Replacing symbol "{symbol}" with "{replacement}".')
    assert len(symbol) == len(replacement), "symbol and replacement must have same length"
    assert len(symbol) != 3 or symbol[0:2] == replacement[0:2], "column symbols may only differ in the column name"

    # step 1: find all matches
    for anchor, tag, context, container, mapping in find(model, symbol):
        # step 2: replace symbol in each matching mapping found in model

        # case: mapping is a `str` column name in a vizcol context _or_ a source-defs columns `list`
        if isinstance(mapping, str) and isinstance(container, list):
            assert mapping == symbol[-1], "expected mapping to match the column name"
            for idx, val in enumerate(container):
                if val == mapping:
                    container[idx] = replacement[-1]

        # case: mapping is a `dict` pseudo-column and `source` is a `str` column name
        elif isinstance(mapping, dict) and isinstance(mapping.get('source'), str):
            assert mapping['source'] == symbol[-1], "expected pseudo-column `source` to match the column name"
            mapping['source'] = replacement[-1]

        # case: mapping is a `dict` pseudo-column and `source` is a source path `list`
        elif isinstance(mapping, dict) and isinstance(mapping.get('source'), list):
            _replace_symbol_in_source_path(anchor, mapping['source'], symbol, replacement)

        # case: mapping is a `str` sourcekey in a source-defs source `dict`
        elif isinstance(mapping, str) and isinstance(container, dict):
            assert mapping in container, "expected to find sourcekey `mapping` in sources `container`"
            assert isinstance(container[mapping], dict), "expected `container[mapping]` to be a source definition dict"
            _replace_symbol_in_source_path(anchor, container[mapping]['source'], symbol, replacement)

        # case: mapping is a `list` constraint name in a vizcol or vizfkey context _or_ a source-defs fkeys `list`
        elif isinstance(mapping, list) and isinstance(container, list):
            assert _is_constraint_match(mapping, symbol), "expected mapping to match the constraint name"
            for idx, val in enumerate(container):
                if val == mapping:
                    container[idx] = _rewrite_constraint_name(val, replacement)
        else:
            logger.warning(f'Unhandled case for replace operation')


def prune(model, symbol):
    """Prunes mappings from a model where symbol found in mapping.

    See the 'find' function for notes on what are 'symbols' and how symbols are found in the model.

    **Note on pruning source definitions**
    When this methods prunes a `source-definitions` entry from the model, it will also prune
    `visible-{columns|foreign-keys}` that reference its `sourcekey` directly or as a "path prefix" in a pseudo-column
    definition. It will prune any references found in `wait_for` display attributes. Also, it will recurse over the
    source definitions repeating the above pruning for each sourcekey dependent on the originally affected sourcekey.
    """
    logger.debug(f'Pruning symbol "{symbol}".')
    # step 1: find all matches
    for anchor, tag, context, container, mapping in find(model, symbol):
        # step 2: remove mapping from model, for each mapping found
        if tag in [tags.visible_columns, tags.visible_foreign_keys]:
            container.remove(mapping)
        elif tag == tags.source_definitions:
            logger.debug(f'Removing "{tag}" mapping "{mapping}" from container "{container}".')
            if isinstance(container, list):
                # step 2.a. match found in 'columns' or 'fkeys' lists, remove item and continue
                container.remove(mapping)
            else:
                # step 2.b. match found in 'sources' dictionary, remove then search for dependencies
                assert isinstance(container, dict), "Expected dictionary typed container"
                del container[mapping]

                # step 2.c. find all dependencies on sourcekey in anchor's annotations and remove them
                for match in _find_sourcekey(anchor, mapping):
                    logger.debug(f'Removing "{mapping}" sourcekey dependency; mapping "{match.mapping}" from "{match.tag}".')
                    if match.tag in [tags.visible_columns, tags.visible_foreign_keys]:
                        match.container.remove(match.mapping)
                    elif match.tag == tags.citation:
                        del match.anchor.annotations[match.tag]
                    elif match.tag == tags.source_definitions:
                        del match.container[match.mapping]
                    else:
                        logger.warning(f'Unexpected tag "{match.tag}".')
        else:
            logger.warning(f'Unhandled tag "{tag}".')


def find(model, symbol):
    """Finds mappings within a model where the mapping contains a given symbol.

    Searches the following annotation tags:
    - source-definitions
    - visible-columns
    - visible-foreign-keys

    Presently, there are two forms of symbols:
    - constraint: `[schema_name, constraint_name]` may refer to a key or fkey. If `constraint_name` is None then the
                  match is based only on `schema_name`
    - column: `[schema_name, table_name, column_name]`

    returns: list containing Match(anchor, tag, context, container, mapping) tuples
    - anchor: the model object that anchors the mapping
    - tag: the annotation tag where the mapping was found
    - context: the annotation context where the mapping was found
    - container: the parent container of the mapping
    - mapping: the mapping in which the symbol was found
    """
    # [NOTE] If/when table must be supported, the ambiguity could be addressed as:
    # - table [schema_name, table_name, None] where the final None in the column category implies that we are removing
    #   not a single column but the whole table (and hence all of its columns)
    matches = []

    # At present, mappings only reside in model elements: table and column.
    for schema in model.schemas.values():
        for table in schema.tables.values():
            for tag in table.annotations:

                # case: visible-columns or visible-foreign-keys
                if tag == tags.visible_columns or tag == tags.visible_foreign_keys:
                    for context in table.annotations[tag]:
                        if context == 'filter':
                            vizsrcs = table.annotations[tag][context].get('and', [])
                        else:
                            vizsrcs = table.annotations[tag][context]

                        for vizsrc in vizsrcs:  # vizsrc is a vizcol or vizfkey entry
                            # case: constraint form of vizsrc
                            if isinstance(vizsrc, list) \
                                    and _is_constraint_match(vizsrc, symbol):
                                matches.append(Match(table, tag, context, vizsrcs, vizsrc))
                            # case: pseudo-column form of vizsrc
                            elif isinstance(vizsrc, dict) and 'source' in vizsrc \
                                    and _is_symbol_in_source(table, vizsrc['source'], symbol):
                                matches.append(Match(table, tag, context, vizsrcs, vizsrc))
                            # case: column form of vizsrc
                            elif isinstance(vizsrc, str) \
                                    and [table.schema.name, table.name, vizsrc] == symbol:
                                matches.append(Match(table, tag, context, vizsrcs, vizsrc))

                # case: source-definitions
                elif tag == tags.source_definitions:
                    # search 'columns'
                    cols = table.annotations[tag].get('columns')
                    if isinstance(cols, list) \
                            and len(symbol) == 3 \
                            and [table.schema.name, table.name] == symbol[0:2] \
                            and symbol[-1] in cols:
                        matches.append(Match(table, tag, 'columns', cols, symbol[-1]))

                    # search 'fkeys'
                    fkeys = table.annotations[tag].get('fkeys')
                    if isinstance(fkeys, list):
                        for fkey in fkeys:
                            if _is_constraint_match(fkey, symbol):
                                matches.append(Match(table, tag, 'fkeys', fkeys, fkey))

                    # search 'sources'
                    sources = table.annotations[tag].get('sources')
                    for sourcekey in sources:
                        if _is_symbol_in_source(table, sources[sourcekey].get('source', []), symbol):
                            matches.append(Match(table, tag, 'sources', sources, sourcekey))

                    # search 'search-box'
                    search_box = table.annotations[tag].get(__search_box__)
                    if isinstance(search_box, dict) and isinstance(search_box.get('or'), list):
                        for search_col in search_box['or']:
                            if _is_symbol_in_source(table, search_col.get('source'), symbol):
                                matches.append(Match(table, tag, __search_box__, search_box['or'], search_col))

    return matches


def _is_constraint_match(constraint_name, symbol):
    """Tests if the constraint name matches the given symbol.

    :param constraint_name: a constraint name pair
    :param symbol: either a constraint name part or a schema name
    """
    if not (isinstance(constraint_name, list) and len(constraint_name) == 2):
        # case: malformed constraint name
        return False
    if not isinstance(symbol, list) or len(symbol) != 2:
        # case: symbol not a constraint name
        return False
    elif symbol[1]:
        # case: symbol is a complete constraint name
        return constraint_name == symbol
    else:
        # case: symbol is a schema name only
        return constraint_name[0] == symbol[0]


def _is_symbol_in_source(table, source, symbol):
    """Finds symbol in a source mapping.
    """

    # case: source is a column name
    if isinstance(source, str):
        return [table.schema.name, table.name, source] == symbol

    # case: source is a path, symbol is a constraint
    if isinstance(source, list) and isinstance(symbol, list):

        # case: symbol is a constraint
        if len(symbol) == 2:
            for pathelem in source:
                if isinstance(pathelem, dict):
                    constraint_name = pathelem.get('inbound') or pathelem.get('outbound')
                    if _is_constraint_match(constraint_name, symbol):
                        return True

        # case: symbol is a column name
        #  -- start with column in path and test if matches last value of source path
        #  -- then determine if the table reference matches (last fkey in/out points to correct table)
        #  -- isinstance(source[-1], str) and source[-1] == symbol[-1]
        #  -- isinstance(source[-2], dict) -- ie, a constraint
        #  -- lookup fkey = model.fkeys(*source[-2].get('inbound' or 'outbound')
        #  -- if 'inbound' and fkey.table == symbol's table
        #  -- if 'outbound' and fkey.pk_table == symbol's table
        if len(symbol) == 3 \
            and len(source) >= 2 \
            and source[-1] == symbol[-1] \
            and isinstance(source[-2], dict):

            # case: inbound fkey
            if 'inbound' in source[-2]:
                fkey = table.schema.model.fkey(source[-2]['inbound'])
                return [fkey.table.schema.name, fkey.table.name] == symbol[0:2]

            # case: outbound fkey
            elif 'outbound' in source[-2]:
                fkey = table.schema.model.fkey(source[-2]['outbound'])
                return [fkey.pk_table.schema.name, fkey.pk_table.name] == symbol[0:2]

    return False


def _find_sourcekey(table, sourcekey):
    """Find usages of `sourcekey` in `table`'s annotations.
    """
    matches = []

    # case: source-definitions
    #                           sources, <sourcekey>, display, wait_for (markdown_pattern)
    #                                                 source, [0], sourcekey
    #
    # case: citation
    #                 wait_for (*_pattern)
    #
    # case: visible-columns or visible-foreign-keys
    #                        <context>, <pseudocolumn>, sourcekey
    #                                                   source, [0], sourcekey
    #                                                   display, wait_for (markdown_pattern)

    # step 1: find all references to sourcekey in source-definitions (recursively)
    sources = table.annotations[tags.source_definitions].get('sources', {})
    sourcekeys = _find_dependent_sourcekeys(sourcekey, sources)
    matches.extend([
        Match(table, tags.source_definitions, None, sources, sourcekey)
        for sourcekey in sourcekeys
    ])

    # step 2: find all other references to the set of {sourcekeys} | {sourcekey} found in first step
    sourcekeys.add(sourcekey)

    #      2.a.: find sourcekey in citations
    citation = table.annotations.get(tags.citation)
    if citation:
        for sourcekey in sourcekeys:
            if sourcekey in citation.get('wait_for', []):
                matches.append(Match(table, tags.citation, None, None, None))
                break

    #      2.b.: find sourcekey in visible-columns or visible-foreign-keys
    for tag in (tags.visible_columns, tags.visible_foreign_keys):
        for context in table.annotations.get(tag, {}):
            vizcols = table.annotations[tag][context]
            for vizcol in vizcols:
                for sourcekey in sourcekeys:
                    if _is_dependent_on_sourcekey(sourcekey, vizcol):
                        matches.append(Match(table, tag, context, vizcols, vizcol))

    return matches


def _find_dependent_sourcekeys(sourcekey, sources, deps=None):
    """Find 'sourcekey' dependencies in 'sources' source definitions.
    """
    # initialize deps to empty set
    deps = deps or set()

    # remove self from sources
    if sourcekey in sources:
        sources = sources.copy()
        sources.pop(sourcekey)

    # test if each 'candidate' source is dependent on this 'sourcekey'
    for candidate in sources:
        # case: source 'search-box', need to check sources in its `or` block, but do not recurse
        if candidate == __search_box__:
            for source_def in sources[__search_box__].get('or', []):
                if _is_dependent_on_sourcekey(sourcekey, source_def):
                    # add 'search-box' to deps, continue
                    deps.add(__search_box__)
                    break
        # case: all other sources
        if _is_dependent_on_sourcekey(sourcekey, sources[candidate]):
            # add to deps
            deps.add(candidate)
            # union with deps of the candidate
            deps.union(_find_dependent_sourcekeys(candidate, sources, deps=deps))

    return deps


def _is_dependent_on_sourcekey(sourcekey, source_def):
    """Tests if 'source_def' is dependent on 'sourcekey'.
    """
    # case: source_def is not a pseudo-column
    if not isinstance(source_def, dict):
        return False
    # case: sourcekey referenced directly
    if sourcekey == source_def.get('sourcekey'):
        return True
    # case: sourcekey in path prefix
    if not isinstance(source_def.get('source'), str) and sourcekey == source_def.get('source', [{}])[0].get('sourcekey'):
        return True
    # case: sourcekey in wait_for
    if sourcekey in source_def.get('display', {}).get('wait_for', []):
        return True
    # not dependent
    return False


def _rewrite_constraint_name(constraint_name, replacement):
    """Rewrites constraint name according to replacement symbol.
    """
    return [
        replacement[0],
        replacement[1] if replacement[1] else constraint_name[1]
    ]


def _replace_symbol_in_source_path(anchor, source, symbol, replacement):
    """Replaces `symbol` with `replacement` in `source`.
    """
    assert isinstance(source, list), 'expected source to be a list'
    assert source, 'expected source to be non-empty'

    if len(symbol) == 3:  # case: column symbol replacement
        assert source[-1] == symbol[-1], "expected source path to terminate in symbol column name"
        source[-1] = replacement[-1]

    if len(symbol) == 2:  # case: constraint name replacement
        assert _is_symbol_in_source(anchor, source, symbol), "expected to find symbol in mapping source"
        for i, pathelem in enumerate(source):
            if not isinstance(pathelem, dict):
                continue
            for direction in ('inbound', 'outbound'):
                if _is_constraint_match(pathelem.get(direction), symbol):
                    pathelem[direction] = _rewrite_constraint_name(pathelem[direction], replacement)
                    break
